/**
 ** locoDemoMotionTaskCodels.c
 **
 ** Codels called by execution task locoDemoMotionTask
 **
 ** Author: 
 ** Date: 
 **
 **/

#include <math.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "server/locoDemoHeader.h"

/* Inclusion specifiques */
#include "locoDemoPom.h"
#include "locoDemoConst.h"

/* Pour la simulation des deplacements:
   Memorisation des vitesses simulees du robot */
static struct {
  double v;
  double w;
}speedSimul;

/*-------------------- PROTOTYPES DES FONCTIONS LOCALES --------------------*/

static STATUS locoDemoComputConfig   (double  v, double w, int *bilan);
static STATUS locoDemoPositionControl(double *vRef, double *wRef, int *bilan);
static STATUS locoDemoErrorsGet      (double *pdLongit, double *pdTransv);
static STATUS locoDemoVelocityGet    (double *v, double *w, int *bilan);
static STATUS locoDemoVelocitySet    (double  v, double  w, int *bilan);


/*------------------------------------------------------------------------
 *
 * initOdoAndAsserv  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
initOdoAndAsserv(int *report)
{
	/* Pointeurs sur les structures a initialiser */
	GENPOS_CART_STATE    *configId;
	GENPOS_CART_REF       *refId;
	GENPOS_CART_CMD_PARAM *cmdId;
	
	/* Initialisation de la configuration */
	configId = &locoDemoDataStrId->robot;
	configId->xRob = 0.;
	configId->yRob = 0.;
	configId->theta = 0.;
	
	/* Initialisation des vitesses */
	configId->v = 0.;
	configId->w = 0.;
	
	/* Initialisation des consignes */
	refId = &locoDemoDataStrId->ref;
	refId->dataType = GENPOS_POS_DATA;
	refId->backFlag = GENPOS_FORWARD_MOTION;
	refId->x = 0.;
	refId->y = 0.;
	refId->v = 0.;
	refId->w = 0.;
	
	refId->vmax = LOCODEMO_VMAX;
	refId->wmax = LOCODEMO_WMAX;
	refId->linAccelMax = LOCODEMO_ACCEL_LIN_MAX;
	refId->angAccelMax = LOCODEMO_ACCEL_ANG_MAX;

	/* Initialisation des parametres de commandes */
	cmdId = &locoDemoDataStrId->cmd;
	cmdId->KpLongit = LOCODEMO_KP_LONGIT;
	cmdId->KpTransv = LOCODEMO_KP_TRANSV;
	
	/* Initialize Pom */
	initPomPosters();
	
	return(OK);
}

/*------------------------------------------------------------------------
 *
 * odoAndAsserv  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 *              S_locoDemo_EMERGENCY_STOP
 * 
 * Returns:    OK or ERROR
 */

STATUS
odoAndAsserv(int *report)
{
	GENPOS_CART_REF *refId = &locoDemoDataStrId->ref;
	double v, w;   
	double vRef, wRef;
	
	/*
	 * Obtention des vitesses du robot
	 */
	if (locoDemoVelocityGet (&v, &w, report)  != OK)
		return(ERROR);
	
	/*
	 * Calcul de la position
	 */
	if (locoDemoComputConfig (v, w, report) != OK)
		return(ERROR);
	
	switch (refId->dataType) {
	case GENPOS_POS_DATA:
		/* 
		 * Calcul des vitesses d'asservissement
		 */
		if (locoDemoPositionControl (&vRef, &wRef, report) != OK)
			return(ERROR);
		break;
	case GENPOS_SPEED_DATA:
		vRef = (refId->backFlag == GENPOS_FORWARD_MOTION ? refId->v :
		    -refId->v);
		wRef = refId->w;
		break;
	default:
		errnoSet(S_locoDemo_BAD_REF);
		return ERROR;
	}
	/* 
	 * Calcul des vitesses d'asservissement
	 */
	if (locoDemoVelocitySet (vRef, wRef, report) != OK)
		return(ERROR);
	
	updatePomPosters();
	
	return(OK);
}
/*------------------------------------------------------------------------
 * SurvLine
 *
 * Description: 
 *
 * Reports:      OK
 */

/* survLine  -  codel EXEC of SurvLine
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
survLine(SURV_LINE_STR *survLineStr, int *report)
{
	double xLine, yLine;       /* Vecteur directeur de la droite */
	double nLine;              /* Norme du vecteur */
	double xPtSeg, yPtSeg;     /* Vecteur 1er point- pt surveille */
	double distPtLine;         /* Distance point-ligne */
	
	/* Vecteur directeur de la droite */
	xLine =  survLineStr->pt2[0] - survLineStr->pt1[0];
	yLine =  survLineStr->pt2[1] - survLineStr->pt1[1];
	nLine = sqrt(xLine * xLine + yLine * yLine);
	
	/* Vecteur 1er point - robot */
	xPtSeg = locoDemoDataStrId->robot.xRob - survLineStr->pt1[0];
	yPtSeg = locoDemoDataStrId->robot.yRob - survLineStr->pt1[1];
	
	/* Distance orientee point surveille-ligne:
	   pas passee: negative,  passe: positive */
	distPtLine = (xLine * yPtSeg - yLine * xPtSeg)/nLine;
	
	/* Point surveille au dessus de la ligne */
	if (distPtLine > 0)
		return(END);
	
	/* En dessous de la ligne */
	else
		return(EXEC);
}


/*------ FONCTIONS AUXILIAIRES de odoAndAsserv ---------------------------*/



/*****************************************************************************
 *
 *  locoDemoComputConfig  -  Calcul la configuration du robot.
 *
 *  Description: Calcul incremental de la configuration du robot a partir 
 *               des vitesses lineaire et angulaire instantannees.
 *
 *  Retourne: OK ou ERROR
 */

static STATUS locoDemoComputConfig (double v,
				    double w,
				    int *bilan)
{
	double dS, dT;          /* Variations lineaires et angulaires */
	double thetaMoy;        /* Angle moyen entre 2 config successives */
	double c, s;            /* Cos et sin de l'angle moyen */
	double locoDemoPeriod = EXEC_TASK_PERIOD(LOCODEMO_MOTIONTASK_NUM);/* Periode */
	GENPOS_CART_STATE *posId = &(locoDemoDataStrId->robot);
	
	/* Variations lineaires et angulaires */
	dS = v * locoDemoPeriod;
	dT = w * locoDemoPeriod;
	
	/* Vitesses du robot */
	posId->v = v;
	posId->w = w;
	
	/* Configuration du robot */
	thetaMoy = posId->theta + dT / 2;
	c = cos(thetaMoy);
	s = sin(thetaMoy);
	posId->theta = angleLimit(posId->theta + dT);
	posId->xRob += dS * c;
	posId->yRob += dS * s;
	
	return(OK);
}



/*****************************************************************************
 *
 *  locoDemoPositionControl  -  Asservissement en position du robot
 *
 *  Description:
 *  Cette routine permet d'asservir le robot en position et cap
 *
 *  Retourne: OK ou ERROR
 */

static STATUS locoDemoPositionControl (double *vRef,
				       double *wRef,
				       int *bilan)

{
	GENPOS_CART_CMD_PARAM *cmdId = &(locoDemoDataStrId->cmd);
	GENPOS_CART_REF *refId = &(locoDemoDataStrId->ref);
	double dLongit;               /* Erreur de deplacement longitudinal */
	double dTransv;               /* Erreur de deplacement transversal */
	double sense;                 /* Sens de deplacement */
	double v;                     /* Vitesse lineaire de consigne */
	double w;                     /* Vitesse angulaire de consigne */
	double accelLin;              /* Acceleration lineaire */
	double accelAng;              /* Acceleration angulaire */
	static double prevV = 0.;     /* Vitesse linaire a la periode precedente */
	static double prevW = 0.;     /* Vitesse angulaire periode precedente */
	double locoDemoPeriod;
	
	locoDemoPeriod  = EXEC_TASK_PERIOD(LOCODEMO_MOTIONTASK_NUM);
	
	/* Sens de deplacement */
	sense = (refId->backFlag == GENPOS_BACKWARD_MOTION ? -1 : 1);
	
	/* Obtenir les ecarts position/consigne */
	locoDemoErrorsGet (&dLongit, &dTransv);
	
	/* Commande de vitesse lineaire a donner au robot */
	v = refId->v + cmdId->KpLongit*dLongit;
	
	/* Limiter l'acceleration lineaire */
	accelLin = (v - prevV)/locoDemoPeriod;
	if (fabs (accelLin) > refId->linAccelMax)
		v = prevV +  SIGN (accelLin) * refId->linAccelMax * locoDemoPeriod;
	
	/* Limiter la vitesse lineaire */
	if (fabs (v) > refId->vmax)
		v = SIGN (v) * refId->vmax;
	
	/* Memorisation de la vitesse lineaire */
	prevV = v;
	
	/* Commande de la vitesse angulaire */
	w = refId->w + sense * cmdId->KpTransv * dTransv;
	
	/* Limiter l'acceleration angulaire */
	accelAng = (w - prevW)/locoDemoPeriod;
	if (fabs (accelAng) > refId->angAccelMax)
		w = prevW + SIGN (accelAng) * refId->angAccelMax * locoDemoPeriod;
	
	/* Limiter la vitesse angulaire */
	if (fabs (w) >  refId->wmax)
		w = SIGN (w) * refId->wmax;
	
	/* Memorisation de la vitesse angulaire */
	prevW = w;
	
	/* Commande en vitesses */
	*vRef = v;
	*wRef = w;
	
	return(OK);
}


/*****************************************************************************
 *
 *  locoDemoErrorsGet  -  Obtenir les ecarts de position du robot
 *
 *  Description:
 *  Cette fonction permet d'obtenir les ecarts longitudinal et transversal
 *  de la position de consigne dans le repere robot.
 *
 *  Retourne: Neant
 */

static STATUS locoDemoErrorsGet (double *pdLongit,   /* ecart longitudinal */
				 double *pdTransv)   /* ecart transversal */

{
	double dxA, dyA;              /* Erreurs position dans le rep atelier */
	GENPOS_CART_STATE *configId = &(locoDemoDataStrId->robot);
	
	/* Obtenir les erreurs en position dans le rep atelier */
	dxA = locoDemoDataStrId->ref.x - configId->xRob;
	dyA = locoDemoDataStrId->ref.y - configId->yRob;
	
	/* Obtenir les erreurs de deplacement transversal et longitudinal */
	*pdLongit =   dxA * cos(configId->theta) + dyA * sin(configId->theta);
	*pdTransv = - dxA * sin(configId->theta) + dyA * cos(configId->theta);
	
	return(OK);
}



/****************************************************************************
 *
 *  locoDemoVelocityGet  -  Obtention des vitesses du robot.
 *
 *  Description: Simule la lecture odometrique des vitesses 
 *               lineaire et angulaire du robot.
 *
 *  Retourne: OK ou ERROR
 */

static STATUS locoDemoVelocityGet (double *v,    /* Vitesse lin. du robot */
				   double *w,    /* Vitesse ang. du robot */
				   int *bilan) 
{
	*v = speedSimul.v;
	*w = speedSimul.w;
	return(OK);
}



/****************************************************************************
 *
 *  locoDemoVelocitySet  -  Commande en vitesses du robot.
 *
 *  Description: Simule la commande en vitesses du robot.
 *               Hypothese: Le robot suit parfaitement la consigne en 
 *               vitesses (les limites en vitesses et acceleration on deja 
 *               ete limitees lors du calcul de la consigne).
 *
 *  Retourne: OK ou ERROR
 */

static STATUS locoDemoVelocitySet (double v,  /* Vitesse lin. commande' */
				   double w,  /* Vitesse ang. commande' */
				   int *bilan)
{
	speedSimul.v = v;
	speedSimul.w = w;
	return(OK);
}

